@@ -2,18 +2,18 @@
 //
 // FTFont.cc
 //
-// Copyright 2001-2002 Glyph & Cog, LLC
+// Copyright 2001-2003 Glyph & Cog, LLC
 //
 //========================================================================
 
-#ifdef __GNUC__
-#pragma implementation
-#endif
-
 #include <aconf.h>
 
 #if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
 
+#ifdef USE_GCC_PRAGMAS
+#pragma implementation
+#endif
+
 #include <math.h>
 #include <string.h>
 #include "gmem.h"
@@ -48,21 +48,22 @@ FTFontEngine::~FTFontEngine() {
 //------------------------------------------------------------------------
 
 FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
-		       char **fontEnc, GBool pdfFontHasEncoding) {
+		       char **fontEnc, Gushort *codeToGID) {
   char *name;
-  int unicodeCmap, macRomanCmap, msSymbolCmap;
-  int i, j;
+  int i;
 
   ok = gFalse;
   engine = engineA;
   codeMap = NULL;
+  cidToGID = NULL;
+  cidToGIDLen = 0;
+
   if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
     return;
   }
 
   if (!strcmp(face->driver->root.clazz->module_name, "type1") ||
       !strcmp(face->driver->root.clazz->module_name, "cff")) {
-
     mode = ftFontModeCodeMapDirect;
     codeMap = (Guint *)gmalloc(256 * sizeof(Guint));
     for (i = 0; i < 256; ++i) {
@@ -73,66 +74,10 @@ FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
     }
 
   } else {
-
-    // To match up with the Adobe-defined behaviour, we choose a cmap
-    // like this:
-    // 1. If the PDF font has an encoding:
-    //    1a. If the TrueType font has a Microsoft Unicode cmap, use it,
-    //        and use the Unicode indexes, not the char codes.
-    //    1b. If the TrueType font has a Macintosh Roman cmap, use it,
-    //        and reverse map the char names through MacRomanEncoding to
-    //        get char codes.
-    // 2. If the PDF font does not have an encoding:
-    //    2a. If the TrueType font has a Macintosh Roman cmap, use it,
-    //        and use char codes directly.
-    //    2b. If the TrueType font has a Microsoft Symbol cmap, use it,
-    //        and use (0xf000 + char code).
-    // 3. If none of these rules apply, use the first cmap and hope for
-    //    the best (this shouldn't happen).
-    unicodeCmap = macRomanCmap = msSymbolCmap = 0xffff;
-    for (i = 0; i < face->num_charmaps; ++i) {
-      if (face->charmaps[i]->platform_id == 3 &&
-	  face->charmaps[i]->encoding_id == 1) {
-	unicodeCmap = i;
-      } else if (face->charmaps[i]->platform_id == 1 &&
-		 face->charmaps[i]->encoding_id == 0) {
-	macRomanCmap = i;
-      } else if (face->charmaps[i]->platform_id == 3 &&
-		 face->charmaps[i]->encoding_id == 0) {
-	msSymbolCmap = i;
-      }
-    }
-    i = 0;
-    mode = ftFontModeCharCode;
-    charMapOffset = 0;
-    if (pdfFontHasEncoding) {
-      if (unicodeCmap != 0xffff) {
-	i = unicodeCmap;
-	mode = ftFontModeUnicode;
-      } else if (macRomanCmap != 0xffff) {
-	i = macRomanCmap;
-	mode = ftFontModeCodeMap;
-	codeMap = (Guint *)gmalloc(256 * sizeof(Guint));
-	for (j = 0; j < 256; ++j) {
-	  if (fontEnc[j]) {
-	    codeMap[j] = globalParams->getMacRomanCharCode(fontEnc[j]);
-	  } else {
-	    codeMap[j] = 0;
-	  }
-	}
-      }
-    } else {
-      if (macRomanCmap != 0xffff) {
-	i = macRomanCmap;
-	mode = ftFontModeCharCode;
-      } else if (msSymbolCmap != 0xffff) {
-	i = msSymbolCmap;
-	mode = ftFontModeCharCodeOffset;
-	charMapOffset = 0xf000;
-      }
-    }
-    if (FT_Set_Charmap(face, face->charmaps[i])) {
-      return;
+    mode = ftFontModeCodeMapDirect;
+    codeMap = (Guint *)gmalloc(256 * sizeof(Guint));
+    for (i = 0; i < 256; ++i) {
+      codeMap[i] = (int)codeToGID[i];
     }
   }
 
@@ -140,29 +85,75 @@ FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
 }
 
 FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
-		       Gushort *cidToGIDA, int cidToGIDLenA) {
+		       Gushort *cidToGIDA, int cidToGIDLenA, GBool embedded) {
+  int i;
+
   ok = gFalse;
   engine = engineA;
   codeMap = NULL;
+  cidToGID = NULL;
+  cidToGIDLen = 0;
+
   if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
     return;
   }
-  cidToGID = cidToGIDA;
   cidToGIDLen = cidToGIDLenA;
-  mode = ftFontModeCIDToGIDMap;
+  cidToGID = (Gushort *)gmalloc(cidToGIDLen * sizeof(Gushort));
+  memcpy(cidToGID, cidToGIDA, cidToGIDLen * sizeof(Gushort));
+  if (!strcmp(face->driver->root.clazz->module_name, "t1cid")) {
+    mode = ftFontModeCID;
+  } else if (!strcmp(face->driver->root.clazz->module_name, "cff")) {
+    mode = ftFontModeCFFCharset;
+  } else if (embedded) {
+    mode = ftFontModeCIDToGIDMap;
+  } else {
+    mode = ftFontModeUnicode;
+    for (i = 0; i < face->num_charmaps; ++i) {
+      if ((face->charmaps[i]->platform_id == 3 &&
+	   face->charmaps[i]->encoding_id == 1) ||
+	  face->charmaps[i]->platform_id == 0) {
+	break;
+      }
+    }
+    if (i == face->num_charmaps) {
+      i = 0;
+    }
+    FT_Set_Charmap(face, face->charmaps[i]);
+  }
   ok = gTrue;
 }
 
-FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName) {
+FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
+		       GBool embedded) {
+  int i;
+
   ok = gFalse;
   engine = engineA;
   codeMap = NULL;
+  cidToGID = NULL;
+  cidToGIDLen = 0;
+
   if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
     return;
   }
-  cidToGID = NULL;
-  cidToGIDLen = 0;
-  mode = ftFontModeCFFCharset;
+  if (!strcmp(face->driver->root.clazz->module_name, "t1cid")) {
+    mode = ftFontModeCID;
+  } else if (embedded) {
+    mode = ftFontModeCFFCharset;
+  } else {
+    mode = ftFontModeUnicode;
+    for (i = 0; i < face->num_charmaps; ++i) {
+      if ((face->charmaps[i]->platform_id == 3 &&
+	   face->charmaps[i]->encoding_id == 1) ||
+	  face->charmaps[i]->platform_id == 0) {
+	break;
+      }
+    }
+    if (i == face->num_charmaps) {
+      i = 0;
+    }
+    FT_Set_Charmap(face, face->charmaps[i]);
+  }
   ok = gTrue;
 }
 
@@ -173,6 +164,9 @@ FTFontFile::~FTFontFile() {
   if (codeMap) {
     gfree(codeMap);
   }
+  if (cidToGID) {
+    gfree(cidToGID);
+  }
 }
 
 //------------------------------------------------------------------------
@@ -261,15 +255,11 @@ FTFont::FTFont(FTFontFile *fontFileA, double *m) {
     yMax = (int)(1.2 * size);
   }
   // this should be (max - min + 1), but we add some padding to
-  // deal with rounding errors
+  // deal with rounding errors, bogus bboxes, etc.
   glyphW = xMax - xMin + 3;
+  glyphW += glyphW >> 1;
   glyphH = yMax - yMin + 3;
-  // another kludge: some CJK TT fonts have bogus bboxes, so add more
-  // padding
-  if (face->num_glyphs > 1000) {
-    glyphW += glyphW >> 1;
-    glyphH += glyphH >> 1;
-  }
+  glyphH += glyphH >> 1;
   if (engine->aa) {
     glyphSize = glyphW * glyphH;
   } else {
@@ -325,7 +315,9 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
   XColor xcolor;
   int bgR, bgG, bgB;
   Gulong colors[5];
-  Guchar *p;
+  Guchar *bitmap, *p;
+  GBool tempBitmap;
+  XImage *img;
   int pix;
   int xOffset, yOffset, x0, y0, x1, y1, gw, gh, w0, h0;
   int xx, yy, xx1;
@@ -338,7 +330,8 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
   }
 
   // generate the glyph pixmap
-  if (!(p = getGlyphPixmap(c, u, &xOffset, &yOffset, &gw, &gh))) {
+  if (!(bitmap = getGlyphPixmap(c, u, &xOffset, &yOffset, &gw, &gh,
+				&tempBitmap))) {
     return gFalse;
   }
 
@@ -360,7 +353,7 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
     w0 = w - x0;
   }
   if (w0 < 0) {
-    return gTrue;
+    goto done;
   }
   if (y0 < 0) {
     y1 = -y0;
@@ -371,17 +364,29 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
     h0 = h - y0;
   }
   if (h0 < 0) {
-    return gTrue;
+    goto done;
+  }
+
+  // getGlyphPixmap may have returned a larger-than-cache-entry
+  // bitmap, in which case we need to allocate a temporary XImage here
+  if (tempBitmap) {
+    if (!(img = XCreateImage(engine->display, engine->visual, engine->depth,
+			     ZPixmap, 0, NULL, gw, gh, 8, 0))) {
+      goto done;
+    }
+    img->data = (char *)gmalloc(gh * img->bytes_per_line);
+  } else {
+    img = image;
   }
 
   // read the X image
   XGetSubImage(engine->display, d, x0, y0, w0, h0, (1 << engine->depth) - 1,
-	       ZPixmap, image, x1, y1);
+	       ZPixmap, img, x1, y1);
 
   if (engine->aa) {
 
     // compute the colors
-    xcolor.pixel = XGetPixel(image, x1 + w0/2, y1 + h0/2);
+    xcolor.pixel = XGetPixel(img, x1 + w0/2, y1 + h0/2);
     XQueryColor(engine->display, engine->colormap, &xcolor);
     bgR = xcolor.red;
     bgG = xcolor.green;
@@ -398,6 +403,7 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
     colors[4] = engine->findColor(r, g, b);
 
     // stuff the glyph pixmap into the X image
+    p = bitmap;
     for (yy = 0; yy < gh; ++yy) {
       for (xx = 0; xx < gw; ++xx) {
 	pix = *p++ & 0xff;
@@ -409,7 +415,7 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
 	  pix = 4;
 	}
 	if (pix > 0) {
-	  XPutPixel(image, xx, yy, colors[pix]);
+	  XPutPixel(img, xx, yy, colors[pix]);
 	}
       }
     }
@@ -420,12 +426,13 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
     colors[1] = engine->findColor(r, g, b);
 
     // stuff the glyph bitmap into the X image
+    p = bitmap;
     for (yy = 0; yy < gh; ++yy) {
       for (xx = 0; xx < gw; xx += 8) {
 	pix = *p++;
 	for (xx1 = xx; xx1 < xx + 8 && xx1 < gw; ++xx1) {
 	  if (pix & 0x80) {
-	    XPutPixel(image, xx1, yy, colors[1]);
+	    XPutPixel(img, xx1, yy, colors[1]);
 	  }
 	  pix <<= 1;
 	}
@@ -435,18 +442,28 @@ GBool FTFont::drawChar(Drawable d, int w, int h, GC gc,
   }
 
   // draw the X image
-  XPutImage(engine->display, d, gc, image, x1, y1, x0, y0, w0, h0);
+  XPutImage(engine->display, d, gc, img, x1, y1, x0, y0, w0, h0);
 
+  if (tempBitmap) {
+    gfree(img->data);
+    img->data = NULL;
+    XDestroyImage(img);
+  }
+ done:
+  if (tempBitmap) {
+    gfree(bitmap);
+  }
   return gTrue;
 }
 
 Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
-			       int *x, int *y, int *w, int *h) {
+			       int *x, int *y, int *w, int *h,
+			       GBool *tempBitmap) {
   FT_GlyphSlot slot;
   FT_UInt idx;
-  int gSize;
+  int rowSize;
   int i, j, k;
-  Guchar *ret;
+  Guchar *ret, *p, *q;
 
   // check the cache
   i = (c & (cacheSets - 1)) * cacheAssoc;
@@ -463,6 +480,7 @@ Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
 	}
       }
       cacheTags[i+j].mru = 0x8000;
+      *tempBitmap = gFalse;
       return cache + (i+j) * glyphSize;
     }
   }
@@ -474,7 +492,9 @@ Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
   idx = getGlyphIndex(c, u);
   // if we have the FT2 bytecode interpreter, autohinting won't be used
 #ifdef TT_CONFIG_OPTION_BYTECODE_INTERPRETER
-  if (FT_Load_Glyph(fontFile->face, idx, FT_LOAD_DEFAULT)) {
+  if (FT_Load_Glyph(fontFile->face, idx,
+		    fontFile->engine->aa ? FT_LOAD_NO_BITMAP
+		                         : FT_LOAD_DEFAULT)) {
     return gFalse;
   }
 #else
@@ -483,8 +503,8 @@ Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
   // anti-aliasing is disabled, this seems to be a tossup - some fonts
   // look better with hinting, some without, so leave hinting on
   if (FT_Load_Glyph(fontFile->face, idx,
-		    fontFile->engine->aa ? FT_LOAD_NO_HINTING
-		                         : FT_LOAD_DEFAULT)) {
+		fontFile->engine->aa ? FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP
+		                     : FT_LOAD_DEFAULT)) {
     return gFalse;
   }
 #endif
@@ -493,38 +513,45 @@ Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
 		                             ft_render_mode_mono)) {
     return gFalse;
   }
+
+  // copy the glyph into the cache or a temporary bitmap
   *x = -slot->bitmap_left;
   *y = slot->bitmap_top;
   *w = slot->bitmap.width;
   *h = slot->bitmap.rows;
-  if (*w > glyphW || *h > glyphH) {
-#if 1 //~ debug
-    fprintf(stderr, "Weird FreeType glyph size: %d > %d or %d > %d\n",
-	    *w, glyphW, *h, glyphH);
-#endif
-    return NULL;
+  if (fontFile->engine->aa) {
+    rowSize = *w;
+  } else {
+    rowSize = (*w + 7) >> 3;
   }
-
-  // store glyph pixmap in cache
-  ret = NULL;
-  for (j = 0; j < cacheAssoc; ++j) {
-    if ((cacheTags[i+j].mru & 0x7fff) == cacheAssoc - 1) {
-      cacheTags[i+j].mru = 0x8000;
-      cacheTags[i+j].code = c;
-      cacheTags[i+j].x = *x;
-      cacheTags[i+j].y = *y;
-      cacheTags[i+j].w = *w;
-      cacheTags[i+j].h = *h;
-      if (fontFile->engine->aa) {
-	gSize = *w * *h;
+  if (*w > glyphW || *h > glyphH) {
+    // the glyph doesn't fit in the bounding box -- return a
+    // temporary, uncached bitmap (this shouldn't happen but some
+    // fonts have incorrect bboxes)
+    ret = (Guchar *)gmalloc(*h * rowSize);
+    *tempBitmap = gTrue;
+  } else {
+    // store glyph pixmap in cache
+    ret = NULL; // make gcc happy
+    for (j = 0; j < cacheAssoc; ++j) {
+      if ((cacheTags[i+j].mru & 0x7fff) == cacheAssoc - 1) {
+	cacheTags[i+j].mru = 0x8000;
+	cacheTags[i+j].code = c;
+	cacheTags[i+j].x = *x;
+	cacheTags[i+j].y = *y;
+	cacheTags[i+j].w = *w;
+	cacheTags[i+j].h = *h;
+	ret = cache + (i+j) * glyphSize;
       } else {
-	gSize = ((*w + 7) >> 3) * *h;
+	++cacheTags[i+j].mru;
       }
-      ret = cache + (i+j) * glyphSize;
-      memcpy(ret, slot->bitmap.buffer, gSize);
-    } else {
-      ++cacheTags[i+j].mru;
     }
+    *tempBitmap = gFalse;
+  }
+  for (k = 0, p = ret, q = slot->bitmap.buffer;
+       k < slot->bitmap.rows;
+       ++k, p += rowSize, q += slot->bitmap.pitch) {
+    memcpy(p, q, rowSize);
   }
   return ret;
 }
@@ -630,20 +657,6 @@ FT_UInt FTFont::getGlyphIndex(CharCode c, Unicode u) {
   case ftFontModeUnicode:
     idx = FT_Get_Char_Index(fontFile->face, (FT_ULong)u);
     break;
-  case ftFontModeCharCode:
-    idx = FT_Get_Char_Index(fontFile->face, (FT_ULong)c);
-    break;
-  case ftFontModeCharCodeOffset:
-    idx = FT_Get_Char_Index(fontFile->face,
-			    (FT_ULong)(c + fontFile->charMapOffset));
-    break;
-  case ftFontModeCodeMap:
-    if (c <= 0xff) {
-      idx = FT_Get_Char_Index(fontFile->face, (FT_ULong)fontFile->codeMap[c]);
-    } else {
-      idx = 0;
-    }
-    break;
   case ftFontModeCodeMapDirect:
     if (c <= 0xff) {
       idx = (FT_UInt)fontFile->codeMap[c];
@@ -664,20 +677,25 @@ FT_UInt FTFont::getGlyphIndex(CharCode c, Unicode u) {
     break;
   case ftFontModeCFFCharset:
 #if 1 //~ cff cid->gid map
+    {
 #if FREETYPE_MAJOR == 2 && FREETYPE_MINOR == 0
-    CFF_Font *cff = (CFF_Font *)((TT_Face)fontFile->face)->extra.data;
+      CFF_Font *cff = (CFF_Font *)((TT_Face)fontFile->face)->extra.data;
 #else
-    CFF_Font cff = (CFF_Font)((TT_Face)fontFile->face)->extra.data;
+      CFF_Font cff = (CFF_Font)((TT_Face)fontFile->face)->extra.data;
 #endif
-    idx = 0;
-    for (j = 0; j < (int)cff->num_glyphs; ++j) {
-      if (cff->charset.sids[j] == c) {
-	idx = j;
-	break;
+      idx = 0;
+      for (j = 0; j < (int)cff->num_glyphs; ++j) {
+	if (cff->charset.sids[j] == c) {
+	  idx = j;
+	  break;
+	}
       }
     }
 #endif
     break;
+  case ftFontModeCID:
+    idx = c;
+    break;
   }
   return idx;
 }
