@@ -4,16 +4,18 @@
 //
 // An X wrapper for the FreeType font rasterizer.
 //
-// Copyright 2001-2002 Glyph & Cog, LLC
+// Copyright 2001-2003 Glyph & Cog, LLC
 //
 //========================================================================
 
 #ifndef FTFONT_H
 #define FTFONT_H
 
+#include <aconf.h>
+
 #if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
 
-#ifdef __GNUC__
+#ifdef USE_GCC_PRAGMAS
 #pragma interface
 #endif
 
@@ -46,12 +48,10 @@ private:
 
 enum FTFontIndexMode {
   ftFontModeUnicode,
-  ftFontModeCharCode,
-  ftFontModeCharCodeOffset,
-  ftFontModeCodeMap,
   ftFontModeCodeMapDirect,
   ftFontModeCIDToGIDMap,
-  ftFontModeCFFCharset
+  ftFontModeCFFCharset,
+  ftFontModeCID
 };
 
 class FTFontFile: public SFontFile {
@@ -59,14 +59,15 @@ public:
 
   // 8-bit font, TrueType or Type 1/1C
   FTFontFile(FTFontEngine *engineA, char *fontFileName,
-	     char **fontEnc, GBool pdfFontHasEncoding);
+	     char **fontEnc, Gushort *codeToGID);
 
   // CID font, TrueType
   FTFontFile(FTFontEngine *engineA, char *fontFileName,
-	     Gushort *cidToGIDA, int cidToGIDLenA);
+	     Gushort *cidToGIDA, int cidToGIDLenA, GBool embedded);
 
   // CID font, Type 0C (CFF)
-  FTFontFile(FTFontEngine *engineA, char *fontFileName);
+  FTFontFile(FTFontEngine *engineA, char *fontFileName,
+	     GBool embedded);
 
   GBool isOk() { return ok; }
   virtual ~FTFontFile();
@@ -76,7 +77,6 @@ private:
   FTFontEngine *engine;
   FT_Face face;
   FTFontIndexMode mode;
-  int charMapOffset;
   Guint *codeMap;
   Gushort *cidToGID;
   int cidToGIDLen;
@@ -107,7 +107,8 @@ public:
 private:
 
   Guchar *getGlyphPixmap(CharCode c, Unicode u,
-			 int *x, int *y, int *w, int *h);
+			 int *x, int *y, int *w, int *h,
+			 GBool *tempBitmap);
   static int charPathMoveTo(FT_Vector *pt, void *state);
   static int charPathLineTo(FT_Vector *pt, void *state);
   static int charPathConicTo(FT_Vector *ctrl, FT_Vector *pt, void *state);
