@@ -4,16 +4,18 @@
 //
 // An X wrapper for the FreeType TrueType font rasterizer.
 //
-// Copyright 2001-2002 Glyph & Cog, LLC
+// Copyright 2001-2003 Glyph & Cog, LLC
 //
 //========================================================================
 
 #ifndef TTFONT_H
 #define TTFONT_H
 
+#include <aconf.h>
+
 #if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
 
-#ifdef __GNUC__
+#ifdef USE_GCC_PRAGMAS
 #pragma interface
 #endif
 
@@ -53,7 +55,6 @@ private:
 enum TTFontIndexMode {
   ttFontModeUnicode,
   ttFontModeCharCode,
-  ttFontModeCharCodeOffset,
   ttFontModeCodeMap,
   ttFontModeCIDToGIDMap
 };
@@ -78,7 +79,6 @@ private:
   TT_Face face;
   TT_CharMap charMap;
   TTFontIndexMode mode;
-  int charMapOffset;
   Guchar *codeMap;
   Gushort *cidToGID;
   int cidToGIDLen;
